@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2002,2003,2007 Martin Hedenfalk <martin@bzero.se>
+ * Copyright (c) 2002-2017  Martin Hedenfalk <martin@bzero.se>
  *
  * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -20,6 +20,9 @@
 
 #include <sys/types.h>
 #include <string.h>
+#ifdef HAVE_STRINGS_H
+# include <strings.h>
+#endif
 #include <stdlib.h>
 #include <assert.h>
 #include <errno.h>
@@ -52,21 +55,18 @@
 #endif
 #define N_(str) str
 
-extern FILE *cfg_yyin;
-extern int cfg_yylex(cfg_t *cfg);
-extern int cfg_lexer_include(cfg_t *cfg, const char *fname);
-extern void cfg_scan_string_begin(const char *buf);
-extern void cfg_scan_string_end(void);
-extern void cfg_scan_fp_begin(FILE *fp);
-extern void cfg_scan_fp_end(void);
-extern char *cfg_qstring;
-
-char *cfg_yylval = 0;
-
 const char confuse_version[] = PACKAGE_VERSION;
 const char confuse_copyright[] = PACKAGE_STRING " by Martin Hedenfalk <martin@bzero.se>";
 const char confuse_author[] = "Martin Hedenfalk <martin@bzero.se>";
 
+char *cfg_yylval = 0;
+
+extern int  cfg_yylex(cfg_t *cfg);
+extern void cfg_yylex_destroy(void);
+extern int  cfg_lexer_include(cfg_t *cfg, const char *fname);
+extern void cfg_scan_fp_begin(FILE *fp);
+extern void cfg_scan_fp_end(void);
+
 static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t *force_opt);
 static void cfg_free_opt_array(cfg_opt_t *opts);
 
@@ -74,10 +74,11 @@ static void cfg_free_opt_array(cfg_opt_t *opts);
 #define STATE_EOF -1
 #define STATE_ERROR 1
 
+#ifndef HAVE_FMEMOPEN
+extern FILE *fmemopen(void *buf, size_t size, const char *type);
+#endif
+
 #ifndef HAVE_STRDUP
-# ifdef HAVE__STRDUP
-#  define strdup _strdup
-# else
 /*
  * Copyright (c) 1988, 1993
  *      The Regents of the University of California.  All rights reserved.
@@ -120,7 +121,6 @@ static char *strdup(const char *s)
 
 	return copy;
 }
-# endif
 #endif
 
 #ifndef HAVE_STRNDUP
@@ -128,9 +128,6 @@ static char *strndup(const char *s, size_t n)
 {
 	char *r;
 
-	if (!s)
-		return NULL;
-
 	r = malloc(n + 1);
 	if (!r)
 		return NULL;
@@ -193,7 +190,7 @@ DLLIMPORT cfg_opt_t *cfg_getopt(cfg_t *cfg, const char *name)
 
 			sec = cfg_getsec(sec, secname);
 			if (!sec) {
-				if (!(cfg->flags & CFGF_IGNORE_UNKNOWN))
+				if (!is_set(CFGF_IGNORE_UNKNOWN, cfg->flags))
 					cfg_error(cfg, _("no such option '%s'"), secname);
 				free(secname);
 				return NULL;
@@ -214,7 +211,7 @@ DLLIMPORT cfg_opt_t *cfg_getopt(cfg_t *cfg, const char *name)
 		}
 	}
 
-	if (!(cfg->flags & CFGF_IGNORE_UNKNOWN))
+	if (!is_set(CFGF_IGNORE_UNKNOWN, cfg->flags))
 		cfg_error(cfg, _("no such option '%s'"), name);
 
 	return NULL;
@@ -253,6 +250,19 @@ DLLIMPORT unsigned int cfg_size(cfg_t *cfg, const char *name)
 	return cfg_opt_size(cfg_getopt(cfg, name));
 }
 
+DLLIMPORT char *cfg_opt_getcomment(cfg_opt_t *opt)
+{
+	if (opt)
+		return opt->comment;
+
+	return NULL;
+}
+
+DLLIMPORT char *cfg_getcomment(cfg_t *cfg, const char *name)
+{
+	return cfg_opt_getcomment(cfg_getopt(cfg, name));
+}
+
 DLLIMPORT signed long cfg_opt_getnint(cfg_opt_t *opt, unsigned int index)
 {
 	if (!opt || opt->type != CFGT_INT) {
@@ -503,15 +513,7 @@ static cfg_opt_t *cfg_dupopt_array(cfg_opt_t *opts)
 
 	return dupopts;
 err:
-	if (dupopts[i].subopts)
-		cfg_free_opt_array(dupopts[i].subopts);
-	if (dupopts[i].def.parsed)
-		free(dupopts[i].def.parsed);
-	if (dupopts[i].def.string)
-		free((void *)dupopts[i].def.string);
-
 	cfg_free_opt_array(dupopts);
-
 	return NULL;
 }
 
@@ -536,6 +538,7 @@ static void cfg_init_defaults(cfg_t *cfg)
 
 	for (i = 0; cfg->opts[i].name; i++) {
 		int j;
+
 		for (j = 0; j < i; ++j) {
 			if (is_set(CFGF_NOCASE, cfg->opts[i].flags | cfg->opts[j].flags)) {
 				if (strcasecmp(cfg->opts[i].name, cfg->opts[j].name))
@@ -562,12 +565,15 @@ static void cfg_init_defaults(cfg_t *cfg)
 			cfg->opts[i].flags |= CFGF_DEFINIT;
 
 			if (is_set(CFGF_LIST, cfg->opts[i].flags) || cfg->opts[i].def.parsed) {
-				int xstate, ret;
+				int xstate, ret = 0;
+				char *buf;
+				FILE *fp;
 
 				/* If it's a list, but no default value was given,
 				 * keep the option uninitialized.
 				 */
-				if (cfg->opts[i].def.parsed == 0 || cfg->opts[i].def.parsed[0] == 0)
+				buf = cfg->opts[i].def.parsed;
+				if (!buf || !buf[0])
 					continue;
 
 				/* setup scanning from the string specified for the
@@ -582,12 +588,25 @@ static void cfg_init_defaults(cfg_t *cfg)
 				else
 					xstate = 2;
 
-				cfg_scan_string_begin(cfg->opts[i].def.parsed);
-				do {
-					ret = cfg_parse_internal(cfg, 1, xstate, &cfg->opts[i]);
-					xstate = -1;
-				} while (ret == STATE_CONTINUE);
-				cfg_scan_string_end();
+				fp = fmemopen(buf, strlen(buf), "r");
+				if (!fp) {
+					/*
+					 * fmemopen() on older GLIBC versions do not accept zero
+					 * length buffers for some reason.  This is a workaround.
+					 */
+					if (strlen(buf) > 0)
+						ret = STATE_ERROR;
+				} else {
+					cfg_scan_fp_begin(fp);
+
+					do {
+						ret = cfg_parse_internal(cfg, 1, xstate, &cfg->opts[i]);
+						xstate = -1;
+					} while (ret == STATE_CONTINUE);
+
+					cfg_scan_fp_end();
+					fclose(fp);
+				}
 
 				if (ret == STATE_ERROR) {
 					/*
@@ -710,8 +729,9 @@ DLLIMPORT cfg_value_t *cfg_setopt(cfg_t *cfg, cfg_opt_t *opt, const char *value)
 				if (!val)
 					return NULL;
 			}
-		} else
+		} else {
 			val = opt->values[0];
+		}
 	}
 
 	switch (opt->type) {
@@ -781,9 +801,10 @@ DLLIMPORT cfg_value_t *cfg_setopt(cfg_t *cfg, cfg_opt_t *opt, const char *value)
 
 	case CFGT_SEC:
 		if (is_set(CFGF_MULTI, opt->flags) || val->section == 0) {
-			if (val->section)
-				val->section->path = 0;
-			cfg_free(val->section);
+			if (val->section) {
+				val->section->path = NULL; /* Global search path */
+				cfg_free(val->section);
+			}
 			val->section = calloc(1, sizeof(cfg_t));
 			if (!val->section)
 				return NULL;
@@ -853,7 +874,7 @@ DLLIMPORT cfg_value_t *cfg_setopt(cfg_t *cfg, cfg_opt_t *opt, const char *value)
 		break;
 
 	default:
-		cfg_error(cfg, "internal error in cfg_setopt(%s, %s)", opt->name, value ?: "NULL");
+		cfg_error(cfg, "internal error in cfg_setopt(%s, %s)", opt->name, (value) ? (value) : "NULL");
 		return NULL;
 	}
 
@@ -938,9 +959,8 @@ DLLIMPORT int cfg_add_searchpath(cfg_t *cfg, const char *dir)
 		return CFG_FAIL;
 	}
 
-	p->next = cfg->path;
-	p->dir = d;
-
+	p->next   = cfg->path;
+	p->dir    = d;
 	cfg->path = p;
 
 	return CFG_SUCCESS;
@@ -1011,7 +1031,7 @@ static int call_function(cfg_t *cfg, cfg_opt_t *opt, cfg_opt_t *funcopt)
 
 static void cfg_handle_deprecated(cfg_t *cfg, cfg_opt_t *opt)
 {
-	if (opt->flags & CFGF_DROP) {
+	if (is_set(CFGF_DROP, opt->flags)) {
 		cfg_error(cfg, _("dropping deprecated configuration option '%s'"), opt->name);
 		cfg_free_value(opt);
 	} else {
@@ -1022,10 +1042,12 @@ static void cfg_handle_deprecated(cfg_t *cfg, cfg_opt_t *opt)
 static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t *force_opt)
 {
 	int state = 0;
+	char *comment = NULL;
 	char *opttitle = NULL;
 	cfg_opt_t *opt = NULL;
 	cfg_value_t *val = NULL;
 	cfg_opt_t funcopt = CFG_STR(0, 0, 0);
+	int ignore = 0;		/* ignore until this token, traverse parser w/o error */
 	int num_values = 0;	/* number of values found for a list option */
 	int rc;
 
@@ -1038,7 +1060,7 @@ static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t
 		int tok = cfg_yylex(cfg);
 
 		if (tok == 0) {
-			/* lexer.l should have called cfg_error */
+			/* lexer.l should have called cfg_error() */
 			goto error;
 		}
 
@@ -1048,35 +1070,58 @@ static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t
 				goto error;
 			}
 
-			if (opt && opt->flags & CFGF_DEPRECATED)
+			if (opt && is_set(CFGF_DEPRECATED, opt->flags))
 				cfg_handle_deprecated(cfg, opt);
 
+			if (comment)
+				free(comment);
+
 			return STATE_EOF;
 		}
 
 		switch (state) {
 		case 0:	/* expecting an option name */
-			if (opt && opt->flags & CFGF_DEPRECATED)
+			if (opt && is_set(CFGF_DEPRECATED, opt->flags))
 				cfg_handle_deprecated(cfg, opt);
 
-			if (tok == '}') {
+			switch (tok) {
+			case '}':
 				if (level == 0) {
 					cfg_error(cfg, _("unexpected closing brace"));
 					goto error;
 				}
+				if (comment)
+					free(comment);
+
 				return STATE_EOF;
-			}
-			if (tok != CFGT_STR) {
+
+			case CFGT_STR:
+				break;
+
+			case CFGT_COMMENT:
+				if (!is_set(CFGF_COMMENTS, cfg->flags))
+					continue;
+
+				if (comment)
+					free(comment);
+				comment = strdup(cfg_yylval);
+				continue;
+
+			default:
 				cfg_error(cfg, _("unexpected token '%s'"), cfg_yylval);
 				goto error;
 			}
+
 			opt = cfg_getopt(cfg, cfg_yylval);
 			if (!opt) {
-				if (cfg->flags & CFGF_IGNORE_UNKNOWN)
-					opt = cfg_getopt(cfg, "__unknown");
-				if (!opt)
-					goto error;
+				if (is_set(CFGF_IGNORE_UNKNOWN, cfg->flags)) {
+					state = 10;
+					break;
+				}
+
+				goto error;
 			}
+
 			if (opt->type == CFGT_SEC) {
 				if (is_set(CFGF_TITLE, opt->flags))
 					state = 6;
@@ -1084,13 +1129,15 @@ static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t
 					state = 5;
 			} else if (opt->type == CFGT_FUNC) {
 				state = 7;
-			} else
+			} else {
 				state = 1;
+			}
 			break;
 
 		case 1:	/* expecting an equal sign or plus-equal sign */
 			if (!opt)
 				goto error;
+
 			if (tok == '+') {
 				if (!is_set(CFGF_LIST, opt->flags)) {
 					cfg_error(cfg, _("attempt to append to non-list option '%s'"), opt->name);
@@ -1109,11 +1156,13 @@ static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t
 				cfg_error(cfg, _("missing equal sign after option '%s'"), opt->name);
 				goto error;
 			}
+
 			if (is_set(CFGF_LIST, opt->flags)) {
 				state = 3;
 				num_values = 0;
-			} else
+			} else {
 				state = 2;
+			}
 			break;
 
 		case 2:	/* expecting an option value */
@@ -1133,13 +1182,20 @@ static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t
 
 			if (cfg_setopt(cfg, opt, cfg_yylval) == 0)
 				goto error;
+
 			if (opt && opt->validcb && (*opt->validcb) (cfg, opt) != 0)
 				goto error;
+
+			/* Inherit last read comment */
+			cfg_opt_setcomment(opt, comment);
+			comment = NULL;
+
 			if (opt && is_set(CFGF_LIST, opt->flags)) {
 				++num_values;
 				state = 4;
-			} else
+			} else {
 				state = 0;
+			}
 			break;
 
 		case 3:	/* expecting an opening brace for a list option */
@@ -1155,14 +1211,15 @@ static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t
 					goto error;
 				++num_values;
 				state = 0;
-			} else
+			} else {
 				state = 2;
+			}
 			break;
 
 		case 4:	/* expecting a separator for a list option, or closing (list) brace */
-			if (tok == ',')
+			if (tok == ',') {
 				state = 2;
-			else if (tok == '}') {
+			} else if (tok == '}') {
 				state = 0;
 				if (opt && opt->validcb && (*opt->validcb) (cfg, opt) != 0)
 					goto error;
@@ -1179,19 +1236,21 @@ static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t
 			}
 
 			val = cfg_setopt(cfg, opt, opttitle);
-			if (opttitle)
-				free(opttitle);
-			opttitle = NULL;
 			if (!val)
 				goto error;
 
-			val->section->path = cfg->path;
+			if (opttitle)
+				free(opttitle);
+			opttitle = NULL;
+
+			val->section->path = cfg->path; /* Remember global search path */
 			val->section->line = cfg->line;
 			val->section->errfunc = cfg->errfunc;
 			rc = cfg_parse_internal(val->section, level + 1, -1, 0);
-			cfg->line = val->section->line;
 			if (rc != STATE_EOF)
 				goto error;
+
+			cfg->line = val->section->line;
 			if (opt && opt->validcb && (*opt->validcb) (cfg, opt) != 0)
 				goto error;
 			state = 0;
@@ -1219,9 +1278,7 @@ static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t
 
 		case 8:	/* expecting a function parameter or a closing paren */
 			if (tok == ')') {
-				int ret = call_function(cfg, opt, &funcopt);
-
-				if (ret != 0)
+				if (call_function(cfg, opt, &funcopt))
 					goto error;
 				state = 0;
 			} else if (tok == CFGT_STR) {
@@ -1242,9 +1299,7 @@ static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t
 
 		case 9:	/* expecting a comma in a function or a closing paren */
 			if (tok == ')') {
-				int ret = call_function(cfg, opt, &funcopt);
-
-				if (ret != 0)
+				if (call_function(cfg, opt, &funcopt))
 					goto error;
 				state = 0;
 			} else if (tok == ',') {
@@ -1255,17 +1310,110 @@ static int cfg_parse_internal(cfg_t *cfg, int level, int force_state, cfg_opt_t
 			}
 			break;
 
+		case 10: /* unknown option, mini-discard parser states: 10-15 */
+			if (comment) {
+				free(comment);
+				comment = NULL;
+			}
+
+			if (tok == '+') {
+				ignore = '=';
+				state = 13; /* Append to list, should be followed by '=' */
+			} else if (tok == '=') {
+				ignore = 0;
+				state = 14; /* Assignment, regular handling */
+			} else if (tok == '(') {
+				ignore = ')';
+				state = 13; /* Function, ignore until end of param list */
+			} else if (tok == '{') {
+				state = 12; /* Section, ignore all until closing brace */
+			} else if (tok == CFGT_STR) {
+				state = 11; /* No '=' ... must be a titled section */
+			} else if (tok == '}' && force_state == 10) {
+				if (comment)
+					free(comment);
+
+				return STATE_CONTINUE;
+			}
+			break;
+
+		case 11: /* unknown option, expecting start of title section */
+			if (tok != '{') {
+				cfg_error(cfg, _("unexpected token '%s'"), cfg_yylval);
+				goto error;
+			}
+			state = 12;
+			break;
+
+		case 12: /* unknown option, recursively ignore entire sub-section */
+			rc = cfg_parse_internal(cfg, level + 1, 10, NULL);
+			if (rc != STATE_CONTINUE)
+				goto error;
+			ignore = '}';
+			state = 13;
+			break;
+
+		case 13: /* unknown option, consume tokens silently until end of func/list */
+			if (tok != ignore)
+				break;
+
+			if (ignore == '=') {
+				ignore = 0;
+				state = 14;
+				break;
+			}
+
+			/* Are we done with recursive ignore of sub-section? */
+			if (force_state == 10) {
+				if (comment)
+					free(comment);
+
+				return STATE_CONTINUE;
+			}
+
+			ignore = 0;
+			state = 0;
+			break;
+
+		case 14: /* unknown option, assuming value or start of list */
+			if (tok == '{') {
+				ignore = '}';
+				state = 13;
+				break;
+			}
+
+			if (tok != CFGT_STR) {
+				cfg_error(cfg, _("unexpected token '%s'"), cfg_yylval);
+				goto error;
+			}
+
+			ignore = 0;
+			if (force_state == 10)
+				state = 15;
+			else
+				state = 0;
+			break;
+
+		case 15: /* unknown option, dummy read of next parameter in sub-section */
+			state = 10;
+			break;
+
 		default:
 			cfg_error(cfg, _("Internal error in cfg_parse_internal(), unknown state %d"), state);
 			goto error;
 		}
 	}
 
+	if (comment)
+		free(comment);
+
 	return STATE_EOF;
 
 error:
 	if (opttitle)
 		free(opttitle);
+	if (comment)
+		free(comment);
 
 	return STATE_ERROR;
 }
@@ -1285,10 +1433,8 @@ DLLIMPORT int cfg_parse_fp(cfg_t *cfg, FILE *fp)
 		return CFG_PARSE_ERROR;
 
 	cfg->line = 1;
-
-	cfg_yyin = fp;
-	cfg_scan_fp_begin(cfg_yyin);
-	ret = cfg_parse_internal(cfg, 0, -1, 0);
+	cfg_scan_fp_begin(fp);
+	ret = cfg_parse_internal(cfg, 0, -1, NULL);
 	cfg_scan_fp_end();
 	if (ret == STATE_ERROR)
 		return CFG_PARSE_ERROR;
@@ -1320,7 +1466,7 @@ static char *cfg_make_fullpath(const char *dir, const char *file)
 	 * if np >= n then the snprintf() was truncated
 	 * (which must be a bug).
 	 */
-	assert(np < len);
+	assert(np < (int)len);
 
 	return path;
 }
@@ -1328,8 +1474,10 @@ static char *cfg_make_fullpath(const char *dir, const char *file)
 DLLIMPORT char *cfg_searchpath(cfg_searchpath_t *p, const char *file)
 {
 	char *fullpath;
+#ifdef HAVE_SYS_STAT_H
 	struct stat st;
 	int err;
+#endif
 
 	if (!p || !file) {
 		errno = EINVAL;
@@ -1357,6 +1505,7 @@ DLLIMPORT char *cfg_searchpath(cfg_searchpath_t *p, const char *file)
 DLLIMPORT int cfg_parse(cfg_t *cfg, const char *filename)
 {
 	int ret;
+	char *fn;
 	FILE *fp;
 
 	if (!cfg || !filename) {
@@ -1364,21 +1513,15 @@ DLLIMPORT int cfg_parse(cfg_t *cfg, const char *filename)
 		return CFG_FILE_ERROR;
 	}
 
-	if (cfg->path) {
-		char *f;
+	if (cfg->path)
+		fn = cfg_searchpath(cfg->path, filename);
+	else
+		fn = cfg_tilde_expand(filename);
+	if (!fn)
+		return CFG_FILE_ERROR;
 
-		if ((f = cfg_searchpath(cfg->path, filename)) == NULL)
-			return CFG_FILE_ERROR;
-
-		free(cfg->filename);
-		cfg->filename = f;
-
-	} else {
-		free(cfg->filename);
-		cfg->filename = cfg_tilde_expand(filename);
-		if (!cfg->filename)
-			return CFG_FILE_ERROR;
-	}
+	free(cfg->filename);
+	cfg->filename = fn;
 
 	fp = fopen(cfg->filename, "r");
 	if (!fp)
@@ -1393,6 +1536,8 @@ DLLIMPORT int cfg_parse(cfg_t *cfg, const char *filename)
 DLLIMPORT int cfg_parse_buf(cfg_t *cfg, const char *buf)
 {
 	int ret;
+	char *fn;
+	FILE *fp;
 
 	if (!cfg) {
 		errno = EINVAL;
@@ -1402,20 +1547,29 @@ DLLIMPORT int cfg_parse_buf(cfg_t *cfg, const char *buf)
 	if (!buf)
 		return CFG_SUCCESS;
 
+	fn = strdup("[buf]");
+	if (!fn)
+		return CFG_PARSE_ERROR;
+
 	free(cfg->filename);
-	cfg->filename = strdup("[buf]");
-	if (!cfg->filename)
-		return CFG_PARSE_ERROR;
+	cfg->filename = fn;
 
-	cfg->line = 1;
-	cfg_scan_string_begin(buf);
-	ret = cfg_parse_internal(cfg, 0, -1, 0);
-	cfg_scan_string_end();
+	fp = fmemopen((void *)buf, strlen(buf), "r");
+	if (!fp) {
+		/*
+		 * fmemopen() on older GLIBC versions do not accept zero
+		 * length buffers for some reason.  This is a workaround.
+		 */
+		if (strlen(buf) > 0)
+			return CFG_FILE_ERROR;
 
-	if (ret == STATE_ERROR)
-		return CFG_PARSE_ERROR;
+		return CFG_SUCCESS;
+	}
 
-	return CFG_SUCCESS;
+	ret = cfg_parse_fp(cfg, fp);
+	fclose(fp);
+
+	return ret;
 }
 
 DLLIMPORT cfg_t *cfg_init(cfg_opt_t *opts, cfg_flag_t flags)
@@ -1507,23 +1661,29 @@ DLLIMPORT int cfg_free_value(cfg_opt_t *opt)
 		return CFG_FAIL;
 	}
 
+	if (opt->comment && !is_set(CFGF_RESET, opt->flags)) {
+		free(opt->comment);
+		opt->comment = NULL;
+	}
+
 	if (opt->values) {
 		unsigned int i;
 
 		for (i = 0; i < opt->nvalues; i++) {
-			if (opt->type == CFGT_STR)
+			if (opt->type == CFGT_STR) {
 				free((void *)opt->values[i]->string);
-			else if (opt->type == CFGT_SEC) {
-				opt->values[i]->section->path = 0;
+			} else if (opt->type == CFGT_SEC) {
+				opt->values[i]->section->path = NULL; /* Global search path */
 				cfg_free(opt->values[i]->section);
-			} else if (opt->type == CFGT_PTR && opt->freecb && opt->values[i]->ptr)
+			} else if (opt->type == CFGT_PTR && opt->freecb && opt->values[i]->ptr) {
 				(opt->freecb) (opt->values[i]->ptr);
+			}
 			free(opt->values[i]);
 		}
 		free(opt->values);
 	}
 
-	opt->values = 0;
+	opt->values  = NULL;
 	opt->nvalues = 0;
 
 	return CFG_SUCCESS;
@@ -1535,11 +1695,13 @@ static void cfg_free_opt_array(cfg_opt_t *opts)
 
 	for (i = 0; opts[i].name; ++i) {
 		free((void *)opts[i].name);
-		if (opts[i].type == CFGT_FUNC || is_set(CFGF_LIST, opts[i].flags))
+		if (opts[i].comment)
+			free(opts[i].comment);
+		if (opts[i].def.parsed)
 			free(opts[i].def.parsed);
-		else if (opts[i].type == CFGT_STR)
+		if (opts[i].def.string)
 			free((void *)opts[i].def.string);
-		else if (opts[i].type == CFGT_SEC)
+		if (opts[i].subopts)
 			cfg_free_opt_array(opts[i].subopts);
 	}
 	free(opts);
@@ -1559,26 +1721,34 @@ static int cfg_free_searchpath(cfg_searchpath_t *p)
 DLLIMPORT int cfg_free(cfg_t *cfg)
 {
 	int i;
+	int isroot = 0;
 
 	if (!cfg) {
 		errno = EINVAL;
 		return CFG_FAIL;
 	}
 
+	if (cfg->comment)
+		free(cfg->comment);
+
 	for (i = 0; cfg->opts[i].name; ++i)
 		cfg_free_value(&cfg->opts[i]);
 
 	cfg_free_opt_array(cfg->opts);
 	cfg_free_searchpath(cfg->path);
 
-	if (cfg->name)
+	if (cfg->name) {
+		isroot = !strcmp(cfg->name, "root");
 		free(cfg->name);
+	}
 	if (cfg->title)
 		free(cfg->title);
 	if (cfg->filename)
 		free(cfg->filename);
 
 	free(cfg);
+	if (isroot)
+		cfg_yylex_destroy();
 
 	return CFG_SUCCESS;
 }
@@ -1626,6 +1796,33 @@ static cfg_value_t *cfg_opt_getval(cfg_opt_t *opt, unsigned int index)
 	return val;
 }
 
+DLLIMPORT int cfg_opt_setcomment(cfg_opt_t *opt, char *comment)
+{
+	char *oldcomment, *newcomment;
+
+	if (!opt || !comment) {
+		errno = EINVAL;
+		return CFG_FAIL;
+	}
+
+	oldcomment = opt->comment;
+	newcomment = strdup(comment);
+	if (!newcomment)
+		return CFG_FAIL;
+
+	if (oldcomment)
+		free(oldcomment);
+	opt->comment = newcomment;
+	opt->flags  |= CFGF_COMMENTS;
+
+	return CFG_SUCCESS;
+}
+
+DLLIMPORT int cfg_setcomment(cfg_t *cfg, const char *name, char *comment)
+{
+	return cfg_opt_setcomment(cfg_getopt(cfg, name), comment);
+}
+
 DLLIMPORT int cfg_opt_setnint(cfg_opt_t *opt, long int value, unsigned int index)
 {
 	cfg_value_t *val;
@@ -1646,7 +1843,13 @@ DLLIMPORT int cfg_opt_setnint(cfg_opt_t *opt, long int value, unsigned int index
 
 DLLIMPORT int cfg_setnint(cfg_t *cfg, const char *name, long int value, unsigned int index)
 {
-	return cfg_opt_setnint(cfg_getopt(cfg, name), value, index);
+	cfg_opt_t *opt;
+
+	opt = cfg_getopt(cfg, name);
+	if (opt && opt->validcb2 && (*opt->validcb2)(cfg, opt, (void *)&value) != 0)
+		return CFG_FAIL;
+
+	return cfg_opt_setnint(opt, value, index);
 }
 
 DLLIMPORT int cfg_setint(cfg_t *cfg, const char *name, long int value)
@@ -1674,7 +1877,13 @@ DLLIMPORT int cfg_opt_setnfloat(cfg_opt_t *opt, double value, unsigned int index
 
 DLLIMPORT int cfg_setnfloat(cfg_t *cfg, const char *name, double value, unsigned int index)
 {
-	return cfg_opt_setnfloat(cfg_getopt(cfg, name), value, index);
+	cfg_opt_t *opt;
+
+	opt = cfg_getopt(cfg, name);
+	if (opt && opt->validcb2 && (*opt->validcb2)(cfg, opt, (void *)&value) != 0)
+		return CFG_FAIL;
+
+	return cfg_opt_setnfloat(opt, value, index);
 }
 
 DLLIMPORT int cfg_setfloat(cfg_t *cfg, const char *name, double value)
@@ -1744,7 +1953,13 @@ DLLIMPORT int cfg_opt_setnstr(cfg_opt_t *opt, const char *value, unsigned int in
 
 DLLIMPORT int cfg_setnstr(cfg_t *cfg, const char *name, const char *value, unsigned int index)
 {
-	return cfg_opt_setnstr(cfg_getopt(cfg, name), value, index);
+	cfg_opt_t *opt;
+
+	opt = cfg_getopt(cfg, name);
+	if (opt && opt->validcb2 && (*opt->validcb2)(cfg, opt, (void *)value) != 0)
+		return CFG_FAIL;
+
+	return cfg_opt_setnstr(opt, value, index);
 }
 
 DLLIMPORT int cfg_setstr(cfg_t *cfg, const char *name, const char *value)
@@ -1821,6 +2036,30 @@ DLLIMPORT int cfg_addlist(cfg_t *cfg, const char *name, unsigned int nvalues, ..
 	return CFG_SUCCESS;
 }
 
+DLLIMPORT cfg_t *cfg_addtsec(cfg_t *cfg, const char *name, const char *title)
+{
+	cfg_opt_t *opt;
+	cfg_value_t *val;
+
+	if (cfg_gettsec(cfg, name, title))
+		return NULL;
+
+	opt = cfg_getopt(cfg, name);
+	if (!opt) {
+		cfg_error(cfg, _("no such option '%s'"), name);
+		return NULL;
+	}
+	val = cfg_setopt(cfg, opt, title);
+	if (!val)
+		return NULL;
+
+	val->section->path = cfg->path; /* Remember global search path. */
+	val->section->line = 1;
+	val->section->errfunc = cfg->errfunc;
+
+	return val->section;
+}
+
 DLLIMPORT int cfg_opt_rmnsec(cfg_opt_t *opt, unsigned int index)
 {
 	unsigned int n;
@@ -1940,6 +2179,7 @@ DLLIMPORT int cfg_opt_nprint_var(cfg_opt_t *opt, unsigned int index, FILE *fp)
 	case CFGT_SEC:
 	case CFGT_FUNC:
 	case CFGT_PTR:
+	case CFGT_COMMENT:
 		break;
 	}
 
@@ -1959,6 +2199,11 @@ DLLIMPORT int cfg_opt_print_indent(cfg_opt_t *opt, FILE *fp, int indent)
 		return CFG_FAIL;
 	}
 
+	if (is_set(CFGF_COMMENTS, opt->flags) && opt->comment) {
+		cfg_indent(fp, indent);
+		fprintf(fp, "/* %s */\n", opt->comment);
+	}
+
 	if (opt->type == CFGT_SEC) {
 		cfg_t *sec;
 		unsigned int i;
@@ -2143,9 +2388,23 @@ DLLIMPORT cfg_validate_callback_t cfg_set_validate_func(cfg_t *cfg, const char *
 	return oldvf;
 }
 
+DLLIMPORT cfg_validate_callback2_t cfg_set_validate_func2(cfg_t *cfg, const char *name, cfg_validate_callback2_t vf)
+{
+	cfg_opt_t *opt;
+	cfg_validate_callback2_t oldvf;
+
+	opt = cfg_getopt_array(cfg->opts, cfg->flags, name);
+	if (!opt)
+		return NULL;
+
+	oldvf = opt->validcb2;
+	opt->validcb2 = vf;
+
+	return oldvf;
+}
+
 /**
  * Local Variables:
- *  version-control: t
  *  indent-tabs-mode: t
  *  c-file-style: "linux"
  * End:
